mod test;
mod opcode;
use opcode::Opcode;
mod input_provider;
use input_provider::{InputProvider, NullProvider};

#[derive(Debug)]
pub struct Computer {
    tape: Vec<i32>,
    ptr_pos: usize,
    input_provider: Box<dyn InputProvider>
}


#[derive(Debug)]
pub struct Error {
    pub msg: String
}

impl Computer {
    fn add(&mut self, left_addy: usize, right_addy: usize, dest_addy: usize) -> Result<(), Error>{
        let memsize = self.tape.len();
        if left_addy < memsize && right_addy < memsize && dest_addy < memsize {
            self.tape[dest_addy] = self.tape[left_addy] + self.tape[right_addy];
            Ok(())
        } else {
            Err(Error{msg: "Bad memory address for add".to_string()})
        }
    }

    fn mult(&mut self, left_addy: usize, right_addy: usize, dest_addy: usize) -> Result<(), Error>{
        let memsize = self.tape.len();
        if left_addy < memsize && right_addy < memsize && dest_addy < memsize {
            self.tape[dest_addy] = self.tape[left_addy] * self.tape[right_addy];
            Ok(())
        } else {
            Err(Error{msg: "Bad memory address for mult".to_string()})
        }
    }

    fn read_and_advance(&mut self) -> Result<i32, Error> {
        if self.ptr_pos < self.tape.len() {
            let res = self.tape[self.ptr_pos];
            self.ptr_pos += 1;
            return Ok(res)
        } else {
            Err(Error{msg: "Read-and-advance exceeded program length".to_string()})
        }
    }

    pub fn from_iter<'a, I>(iter: I) -> Computer
        where I: Iterator<Item = &'a i32> {
        return Computer { 
            tape: iter.map(|item| item.clone()).collect(),
            ptr_pos: 0,
            input_provider: Box::new(NullProvider{})
        }
    }

    pub fn execute(&mut self) -> Result<(), Error> {
        while let Ok(intcode) = self.read_and_advance() {
            match intcode.try_into() {
                Ok(Opcode::ADD) => {
                    let left_address: usize = self.read_and_advance()?.try_into().unwrap();
                    let right_address: usize = self.read_and_advance()?.try_into().unwrap();
                    let dest_address: usize = self.read_and_advance()?.try_into().unwrap();
                    self.add(left_address, right_address, dest_address)?;
                },
                Ok(Opcode::MULT) => {
                    let left_address: usize = self.read_and_advance()?.try_into().unwrap();
                    let right_address: usize = self.read_and_advance()?.try_into().unwrap();
                    let dest_address: usize = self.read_and_advance()?.try_into().unwrap();
                    self.mult(left_address, right_address, dest_address)?;
                },
                Ok(Opcode::TERMINATE) => {
                    return Ok(())
                }
                Err(_) => {
                    println!("Unknown opcode {}", intcode);
                    return Err(Error{msg: "Unknown opcode".to_string()})
                }
            }
        }

        return Err(Error{msg: "Unexpected end-of-program".to_string()})
    }

    pub fn set_position(&mut self, pos: usize, value: i32) {
        self.tape[pos] = value;
    }

    pub fn get_pos(&self, pos: usize) -> &i32 {
        self.tape.get(pos).unwrap()
    }
}